/*
Copyright 2021 The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package cmd

import (
	"encoding/json"
	"fmt"

	"github.com/spf13/cobra"
)

var jsonOutput bool
var verbose bool

type Chain []string

// statsCmd represents the statsDeps command
var statsCmd = &cobra.Command{
	Use:   "stats",
	Short: "Shows metrics about dependency chains",
	Long: `Provides the following metrics:
	1. Total Dependencies: Total number of dependencies of the project
	2. Max Depth of Dependencies: Number of dependencies in the longest dependency chain
	3. Transitive Dependencies: Total number of transitive dependencies (dependencies which are not direct dependencies of the project)`,
	RunE: func(cmd *cobra.Command, args []string) error {
		depGraph, deps, mainModule := getDepInfo()

		// Get all chains starting from main module
		// also get all cycles
		// cycleChains stores the chain containing the cycles and
		// not the actual cycle itself
		var cycleChains []Chain
		var chains []Chain
		var temp Chain
		getChains(mainModule, depGraph, temp, &chains, &cycleChains)

		// get values
		totalDeps := len(deps)
		maxDepth := getMaxDepth(chains)
		directDeps := len(depGraph[mainModule])
		transitiveDeps := totalDeps - directDeps

		if !jsonOutput {
			fmt.Printf("Total Dependencies: %d \n", totalDeps)
			fmt.Printf("Max Depth Of Dependencies: %d \n", maxDepth)
			fmt.Printf("Transitive Dependencies: %d \n", transitiveDeps)
		}

		if verbose {
			fmt.Println("All dependencies:")
			printDeps(deps)
		}

		// print the longest chain
		if verbose {
			fmt.Println("Longest chain/s: ")
			for _, chain := range getLongestChains(maxDepth, chains) {
				printChain(chain)
			}
		}

		if jsonOutput {
			// create json
			outputObj := struct {
				TotalDeps int `json:"totalDependencies"`
				MaxDepth  int `json:"maxDepthOfDependencies"`
				TransDeps int `json:"transitiveDependencies"`
			}{
				TotalDeps: totalDeps,
				MaxDepth:  maxDepth,
				TransDeps: transitiveDeps,
			}
			outputRaw, err := json.MarshalIndent(outputObj, "", "\t")
			if err != nil {
				return err
			}
			fmt.Print(string(outputRaw))
		}
		return nil
	},
}

// get the length of the longest dependency chain
func getMaxDepth(chains []Chain) int {
	maxDeps := 0
	for _, chain := range chains {
		maxDeps = max(maxDeps, len(chain))
	}
	// for A -> B -> C the depth is 3
	return maxDeps
}

func getLongestChains(maxDepth int, chains []Chain) []Chain {
	var longestChains []Chain
	for _, chain := range chains {
		if len(chain) == maxDepth {
			longestChains = append(longestChains, chain)
		}
	}
	return longestChains
}

func init() {
	rootCmd.AddCommand(statsCmd)
	statsCmd.Flags().BoolVarP(&verbose, "verbose", "v", false, "Get additional details")
	statsCmd.Flags().BoolVarP(&jsonOutput, "json", "j", false, "Get the output in JSON format")
}
